/* -*- C -*-
 * $RCSfile: PgSQL.c,v $
 * $Date: 2001/03/26 20:40:38 $
 * $Revision: 1.1.1.1 $
 * Copyright (C) 2001 Timothy H. Keitt
 */

#include <R.h>
#include <Rinternals.h>
#include <libpq-fe.h>

static SEXP
mkInteger(int val) {
  SEXP out;
  PROTECT(out = allocVector(INTSXP, 1));
  INTEGER(out)[0] = val;
  UNPROTECT(1);
  return(out);
}

static SEXP
mkLogical(int val) {
  SEXP out;
  PROTECT(out = allocVector(LGLSXP, 1));
  LOGICAL(out)[0] = val;
  UNPROTECT(1);
  return(out);
}

static SEXP
mkString_safe(char *string) {
  if (string == NULL) string = "";
  return(mkString(string));
}

void
PgSQLcloseConnection(SEXP connPtr) {

  PGconn *conn = (PGconn *) R_ExternalPtrAddr(connPtr);
  if (PQstatus(conn) == CONNECTION_OK) PQfinish(conn);
  return;

}

SEXP
PgSQLconnect(SEXP connInfo) {

  PGconn *conn;
  SEXP connPtr;

  char *connInfoStr = CHAR(STRING_ELT(connInfo, 0));

  PROTECT(connInfo = coerceVector(connInfo, STRSXP));
  conn = PQconnectdb(connInfoStr);
  UNPROTECT(1);

  PROTECT(connPtr = R_MakeExternalPtr((void *) conn,
				      mkChar("PgSQL connection"),
				      mkInteger(1)));
  UNPROTECT(1);

  return(connPtr);

}

SEXP
PgSQLconnectionOK(SEXP connPtr) {

  PGconn *conn = (PGconn *) R_ExternalPtrAddr(connPtr);

  int status = PQstatus(conn);

  if (status != CONNECTION_OK) return(mkLogical(0));

  return(mkLogical(1));

}

SEXP
PgSQLconnectionInfo(SEXP connPtr) {

  SEXP connInfo, listNames;
  PGconn *conn = (PGconn *) R_ExternalPtrAddr(connPtr);

  int status = PQstatus(conn);

  if (status != CONNECTION_OK) return(R_NilValue);

  PROTECT(connInfo = allocVector(VECSXP, 12));
  PROTECT(listNames = allocVector(STRSXP, 12));

  SET_VECTOR_ELT(connInfo, 0, mkInteger(status));
  SET_VECTOR_ELT(listNames, 0, mkChar("connection.status")); 

  SET_VECTOR_ELT(connInfo, 1, mkString_safe(PQerrorMessage(conn)));
  SET_VECTOR_ELT(listNames, 1, mkChar("last.message")); 

  SET_VECTOR_ELT(connInfo, 2, mkString_safe(PQdb(conn)));
  SET_VECTOR_ELT(listNames, 2, mkChar("database.name")); 

  SET_VECTOR_ELT(connInfo, 3, mkString_safe(PQhost(conn)));
  SET_VECTOR_ELT(listNames, 3, mkChar("host.name")); 

  SET_VECTOR_ELT(connInfo, 4, mkString_safe(PQoptions(conn)));
  SET_VECTOR_ELT(listNames, 4, mkChar("connection.options")); 

  SET_VECTOR_ELT(connInfo, 5, mkString_safe(PQuser(conn)));
  SET_VECTOR_ELT(listNames, 5, mkChar("user.name")); 

  SET_VECTOR_ELT(connInfo, 6, mkString_safe(PQpass(conn)));
  SET_VECTOR_ELT(listNames, 6, mkChar("connection.password")); 

  SET_VECTOR_ELT(connInfo, 7, mkString_safe(PQport(conn)));
  SET_VECTOR_ELT(listNames, 7, mkChar("connection.port")); 

  SET_VECTOR_ELT(connInfo, 8, mkString_safe(PQtty(conn)));
  SET_VECTOR_ELT(listNames, 8, mkChar("connection.tty")); 

  SET_VECTOR_ELT(connInfo, 9, mkInteger(PQsocket(conn)));
  SET_VECTOR_ELT(listNames, 9, mkChar("connection.socket")); 

  SET_VECTOR_ELT(connInfo, 10, mkInteger(PQclientEncoding(conn)));
  SET_VECTOR_ELT(listNames, 10, mkChar("client.encoding")); 

  SET_VECTOR_ELT(connInfo, 11, mkInteger(PQbackendPID(conn)));
  SET_VECTOR_ELT(listNames, 11, mkChar("backend.pid")); 

  setAttrib(connInfo, R_NamesSymbol, listNames);

  UNPROTECT(2);
  return(connInfo);

}

static SEXP
PgSQLgetString(PGresult *result, unsigned row, unsigned col) {

  char *value;

  if (PQgetisnull(result, row, col)) return(NA_STRING);

  value = PQgetvalue(result, row, col);

  return(mkChar(value));

}

static int
PgSQLgetInteger(PGresult *result, unsigned row, unsigned col) {

  int out, warn = 0;
  char *value;

  if (PQgetisnull(result, row, col)) return(NA_INTEGER);

  value = PQgetvalue(result, row, col);
  
  out = IntegerFromString(mkChar(value), &warn);

  CoercionWarning(warn);

  return(out);

}

static double
PgSQLgetDouble(PGresult *result, unsigned row, unsigned col) {

  double out;
  int warn = 0;
  char *value;

  if (PQgetisnull(result, row, col)) return(NA_INTEGER);

  value = PQgetvalue(result, row, col);
  
  out = RealFromString(mkChar(value), &warn);

  CoercionWarning(warn);

  return(out);

}

static int
PgSQLgetLogical(PGresult *result, unsigned row, unsigned col) {

  if (PQgetisnull(result, row, col)) return(NA_LOGICAL);

  if (PQgetvalue(result, row, col) == "t") return(1);

  return(0);
}

static SEXP
PgSQLgetTypeName(PGconn *conn, unsigned type) {

  PGresult *result;
  char query[80];

  sprintf(query, "select typname from pg_type where oid = %d\0", type);

  result = PQexec(conn, query);

  if (PQresultStatus(result) != PGRES_TUPLES_OK) return(NA_STRING);

  return(PgSQLgetString(result, 0, 0));

}

static SEXP
PgSQLfetchResult(PGconn *conn, PGresult *result) {

  SEXP tupleFrame, column, colNames;
  SEXP colTypes, typeNames, rowNames;
  unsigned row, col, rows, cols, type;

  rows = PQntuples(result);
  cols = PQnfields(result);

  PROTECT(tupleFrame = allocVector(VECSXP, cols));
  PROTECT(colNames = allocVector(STRSXP, cols));
  PROTECT(colTypes = allocVector(INTSXP, cols));
  PROTECT(typeNames = allocVector(STRSXP, cols));
  
  for (col = 0; col < cols; ++col) {

    type = PQftype(result, col);

    INTEGER(colTypes)[col] = type;
    SET_VECTOR_ELT(typeNames, col, PgSQLgetTypeName(conn, type));
    SET_VECTOR_ELT(colNames, col, mkChar(PQfname(result, col)));

    switch (type) {

    case 16:			/* bool */

      PROTECT(column = allocVector(LGLSXP, rows));
    
      for (row = 0; row < rows; ++row)
	LOGICAL(column)[row] = PgSQLgetLogical(result, row, col);

      break;

    case 20:			/* int8 */
    case 21:			/* int2 */
    case 23:			/* int4 */
    case 26:			/* oid */

      PROTECT(column = allocVector(INTSXP, rows));
    
      for (row = 0; row < rows; ++row)
	INTEGER(column)[row] = PgSQLgetInteger(result, row, col);

      break;

    case 700:			/* float4 */
    case 701:			/* float8 */
    case 1700:			/* numeric */

      PROTECT(column = allocVector(REALSXP, rows));
    
      for (row = 0; row < rows; ++row)
	REAL(column)[row] = PgSQLgetDouble(result, row, col);

      break;

    default:			/* all other types */

      PROTECT(column = allocVector(STRSXP, rows));
    
      for (row = 0; row < rows; ++row)
	SET_VECTOR_ELT(column, row, PgSQLgetString(result, row, col));

      break;

    }
    
    SET_VECTOR_ELT(tupleFrame, col, column);
    
    UNPROTECT(1);
    
  }
  
  namesgets(tupleFrame, colNames);

  setAttrib(tupleFrame, install("pgsql.column.types"), colTypes);
  setAttrib(tupleFrame, install("pgsql.type.names"), typeNames);
  
  PROTECT(rowNames = allocVector(STRSXP, rows));
  
  for (row = 0; row < rows; ++row)
    SET_VECTOR_ELT(rowNames, row, StringFromInteger(row+1, NULL));
  
  setAttrib(tupleFrame, R_RowNamesSymbol, rowNames);
  
  classgets(tupleFrame, mkString_safe("data.frame"));

  PQclear(result);
  UNPROTECT(5);

  return(tupleFrame);

}

SEXP
PgSQLsendQuery(SEXP connPtr, SEXP queryString) {

  PGresult *result;
  PGconn *conn = (PGconn *) R_ExternalPtrAddr(connPtr);

  if (PQstatus(conn) != CONNECTION_OK)
    error("No database connection");

  PROTECT(queryString = coerceVector(queryString, STRSXP));

  result = PQexec(conn, CHAR(STRING_ELT(queryString, 0)));
  
  UNPROTECT(1);

  switch (PQresultStatus(result)) {

  case PGRES_COMMAND_OK:
  case PGRES_FATAL_ERROR:
  case PGRES_BAD_RESPONSE:

    Rprintf("%s", PQerrorMessage(conn));

  case PGRES_EMPTY_QUERY:
  case PGRES_COPY_OUT:
  case PGRES_COPY_IN:

    PQclear(result);
    return(R_NilValue);
    break;

  case PGRES_NONFATAL_ERROR:
    
    warning(PQerrorMessage(conn));

  case PGRES_TUPLES_OK:

    return(PgSQLfetchResult(conn, result));
    break;

  default:

    PQclear(result);
    error("Query returned an unrecognized status code");
    return(R_NilValue);
    break;

  }

  PQclear(result);
  error("A serious bug exists in PgSQL or your C compiler");
  return(R_NilValue);

}

SEXP
PgSQLappendData(SEXP conn, SEXP name, SEXP data) {



}
